/*
 * $Id: main.cpp,v 1.4 2010-05-22 18:32:48 laci Exp $
 *
 * Project      : RNA motif searching in genomic sequences
 * Description  : the main program
 *
 * Author       : Ladislav Rampasek <rampasek@gmail.com>
 * Institution  : Comenius University in Bratislava
 *
 */

#include <iostream>
#include <string>
#include <iomanip>
#include <fstream>

#include "search.h"

using namespace std;

#ifndef RELEASE
    #define DEF_RELEASE     "dev1.7"
    #define DEF_RELEASEDATE "May 2010"
#endif

static string txt_banner = "RNAMot2 - RNA pattern searching program";
static string txt_usage = "Usage: rnamot2 <descriptor-file> <sequence-file>\n";


int main(int argc, char* argv[]){

    string file_db;		    /* name of sequence file to search           */
    string file_desc;		/* name of file containing descriptor        */
    ifstream dbin, descin;

    int opt_skip_mult = false;	    /* OPTION: True to filter out overlapping hits   */
    int opt_be_quiet = false;		/* OPTION: TRUE to silence verbosity             */
    int opt_print_full = false;     /* OPTION: TRUE to show full alignments          */

  /***********************************************
   * Parse command line arguments
   * TODO: pridat podporu pre options OPT_skip_mult, OPT_be_quiet, OPT_print_full ako v RNABob
   ***********************************************/

    if (argc != 3) die("Incorrect number of arguments\n" + txt_usage);
    file_desc = argv[1];
    file_db = argv[2];

  /***********************************************
   * Initialize sequence file, read descriptor
   * TODO: podpora pre preddefinovane umiestnenia suborov
   ***********************************************/
    /// read descriptor
    descin.open( file_desc.c_str() );
    if(! descin.good()) die("Descriptor file "+file_desc+" could not be opened or it does not exist");
    Descriptor desc(descin);
    if(! desc.is_initialized()) die(desc.error_str());

    /// open sequence file
    dbin.open( file_db.c_str() );
    if(! dbin.good()) die("Sequence file "+file_db+" could not be opened or it does not exist");

  /***********************************************
   * Print header and do the search
   ***********************************************/

    if (! opt_be_quiet)
    {
      cout<<"Starting rnamot2: version "<<DEF_RELEASE<<", "<<DEF_RELEASEDATE<<"\n";
      cout<<"---------------------------------------------------\n";
      cout<<"Database file:                 "<<file_db<<"\n";
      cout<<"Descriptor file:               "<<file_desc<<"\n";
      cout<<"Search order:                  "<<desc.search_order_to_str()<<"\n";
      cout<<"Filter out overlapping hits:   "<<(opt_skip_mult? "yes":"no")<<"\n";
      cout<<"---------------------------------------------------\n";

      cout<<"\n";
      cout<<" seq-f  seq-t     name     description\n";
      cout<<"------ ------ ------------ ------------\n";
    }

    Simple_Search ssearch(desc);
    string line;
    while( get_valuable_line(dbin,line) ){
        if(line[0] != '>') die("Incorrect sequence format");
        line[0]=' ';

        string sq_name, sq_details, sq;
        istringstream sin(line);
        if( !(sin>>sq_name)) die("Incorrect sequence format");
        while( sin>>line ) sq_details+=" "+line;

        while('A'<=dbin.peek() && dbin.peek()<='z') {
            get_valuable_line(dbin,line);
            transform(line.begin(), line.end(), line.begin(), ::toupper);
            sq+=line;
        }
        //cout<<sq_name<<":"<<sq<<endl<<endl;

        int max_motif_length = ssearch.desc->get_max_motif_length();
        int max_seq_length = max(5000, 3*max_motif_length);

        vector< pair<string,int> > seq_blocks;
        int prev_found = 0;
        int found = sq.find("NNNNNNNNNN");
        while(found!=string::npos){
            seq_blocks.push_back(make_pair(sq.substr(prev_found, found-prev_found+1), prev_found));
            while(sq[found+10]=='N') ++found;
            prev_found = found;
            found=sq.find("NNNNNNNNNN",found+1);
        }
        seq_blocks.push_back(make_pair(sq.substr(prev_found, sq.size()), prev_found));

        for(int k=0; k<seq_blocks.size(); k++){
            int pointer = 0;
            vector< pair<string,int> > seq_partitions;
            while(pointer < seq_blocks[k].first.size()){
                int start = max(0, pointer-max_motif_length+1);

                string s = seq_blocks[k].first.substr(start, max_seq_length);
                seq_partitions.push_back(make_pair(s, start));
                pointer = start + max_seq_length;
            }

            for(int j=0; j<seq_partitions.size(); j++){
                ssearch.search(seq_partitions[j].first);

                for(int i=0;i<ssearch.solutions.size();i++){
                    cout<<setw(6)<<ssearch.solutions[i][0].first +1 +seq_partitions[j].second+ seq_blocks[k].second;
                    cout<<" "<<setw(6)<<ssearch.solutions[i].back().second +seq_partitions[j].second+ seq_blocks[k].second;
                    cout<<" "<<setw(12)<<sq_name;
                    cout<<" "<<setw(12)<<sq_details<<endl;
                    cout<<ssearch.solution_to_str(i, seq_partitions[j].first)<<endl;
                }
            }
        }

    }

    return 0;
}
